#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <iomanip>
#include <gdkmm.h>
#include <gtkmm.h>

#include "cfmuvalidate.hh"

CFMUValidate::MainWindow::MainWindow(BaseObjectType* cobject, const Glib::RefPtr<Builder>& builder)
	: Gtk::Window(cobject), m_builder(builder), m_aboutwindow(0),
	  m_textviewresults(0), m_scrolledwindowbrowser(0), m_buttonabout(0), m_buttonvalidate(0), m_buttonquit(0),
	  m_portal_webview(0), m_validate_webview(0),
	  m_portal_webview_close_sighandler(0), m_portal_webview_create_sighandler(0),
	  m_portal_webview_ready_sighandler(0), m_portal_webview_loadstatus_sighandler(0),
	  m_portal_webview_navpolicy_sighandler(0), m_portal_webview_newwndpolicy_sighandler(0),
	  m_portal_webview_rsrcload_sighandler(0), m_portal_webview_consolemsg_sighandler(0),
	  m_validate_webview_close_sighandler(0), m_validate_webview_create_sighandler(0),
	  m_validate_webview_ready_sighandler(0), m_validate_webview_loadstatus_sighandler(0),
	  m_validate_webview_navpolicy_sighandler(0), m_validate_webview_newwndpolicy_sighandler(0),
	  m_validate_webview_rsrcload_sighandler(0), m_validate_webview_consolemsg_sighandler(0),
	  m_app(0), m_wvstate(wvstate_failed)
{
        m_builder->get_widget("cfmuaboutdialog", m_aboutwindow);
	m_builder->get_widget("textviewfplan", m_textviewfplan);
	m_builder->get_widget("textviewresults", m_textviewresults);
	m_builder->get_widget("scrolledwindowbrowser", m_scrolledwindowbrowser);
	m_builder->get_widget("buttonabout", m_buttonabout);
	m_builder->get_widget("buttonvalidate", m_buttonvalidate);
	m_builder->get_widget("buttonquit", m_buttonquit);

	if (m_buttonvalidate)
		m_buttonvalidate->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::validate_clicked));
	if (m_buttonquit)
		m_buttonquit->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::quit_clicked));
	if (m_buttonabout)
		m_buttonabout->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::about_clicked));
	if (m_aboutwindow)
		m_aboutwindow->signal_response().connect(sigc::mem_fun(*this, &MainWindow::about_done));
	if (m_textviewfplan) {
		Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewfplan->get_buffer());
		if (buf) {
			buf->set_text("-(FPL-HBPBX-IG -1P28R/L -SDFGLO/S -LSZH1235 -N0135F070 ZUE -EDNY0026 LSZR -)");
			buf->signal_changed().connect(sigc::mem_fun(*this, &MainWindow::fpl_changed));
		}
	}
	prepare_webview();
}

CFMUValidate::MainWindow::~MainWindow()
{
	remove_portal_sighandlers();
	remove_validate_sighandlers();
	delete m_aboutwindow;
	delete m_portal_webview;
	delete m_validate_webview;
}

void CFMUValidate::MainWindow::add_portal_sighandlers(void)
{
	remove_portal_sighandlers();
	m_portal_webview_close_sighandler = g_signal_connect(m_portal_webview->gobj(), "close-web-view", G_CALLBACK(&MainWindow::close_webview_1), this);
	m_portal_webview_create_sighandler = g_signal_connect(m_portal_webview->gobj(), "create-web-view", G_CALLBACK(&MainWindow::create_webview_1), this);
	m_portal_webview_ready_sighandler = g_signal_connect(m_portal_webview->gobj(), "web-view-ready", G_CALLBACK(&MainWindow::webview_ready_1), this);
	m_portal_webview_loadstatus_sighandler = g_signal_connect(m_portal_webview->gobj(), "notify::load-status", G_CALLBACK(&MainWindow::loadstatus_webview_1), this);
	m_portal_webview_navpolicy_sighandler = g_signal_connect(m_portal_webview->gobj(), "navigation-policy-decision-requested", G_CALLBACK(&MainWindow::policy_decision_required_1), this);
	m_portal_webview_newwndpolicy_sighandler = g_signal_connect(m_portal_webview->gobj(), "new-window-policy-decision-requested", G_CALLBACK(&MainWindow::policy_decision_required_1), this);
	m_portal_webview_rsrcload_sighandler = g_signal_connect(m_portal_webview->gobj(), "resource-load-finished", G_CALLBACK(&MainWindow::resourceload_finished_1), this);
	m_portal_webview_consolemsg_sighandler = g_signal_connect(m_portal_webview->gobj(), "console-message", G_CALLBACK(&MainWindow::console_message_1), this);
}

void CFMUValidate::MainWindow::remove_portal_sighandlers(void)
{
        if (m_portal_webview_close_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_close_sighandler);
		m_portal_webview_close_sighandler = 0;
        }
        if (m_portal_webview_create_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_create_sighandler);
		m_portal_webview_create_sighandler = 0;
        }
	if (m_portal_webview_ready_sighandler) {
		g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_ready_sighandler);
		m_portal_webview_ready_sighandler = 0;
	}
        if (m_portal_webview_loadstatus_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_loadstatus_sighandler);
		m_portal_webview_loadstatus_sighandler = 0;
        }
        if (m_portal_webview_navpolicy_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_navpolicy_sighandler);
		m_portal_webview_navpolicy_sighandler = 0;
        }
        if (m_portal_webview_newwndpolicy_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_newwndpolicy_sighandler);
		m_portal_webview_newwndpolicy_sighandler = 0;
        }
        if (m_portal_webview_rsrcload_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_rsrcload_sighandler);
		m_portal_webview_rsrcload_sighandler = 0;
        }
        if (m_portal_webview_consolemsg_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_consolemsg_sighandler);
		m_portal_webview_consolemsg_sighandler = 0;
        }
}

void CFMUValidate::MainWindow::add_validate_sighandlers(void)
{
	remove_validate_sighandlers();
	m_validate_webview_close_sighandler = g_signal_connect(m_validate_webview->gobj(), "close-web-view", G_CALLBACK(&MainWindow::close_webview_1), this);
	m_validate_webview_create_sighandler = g_signal_connect(m_validate_webview->gobj(), "create-web-view", G_CALLBACK(&MainWindow::create_webview_1), this);
	m_validate_webview_ready_sighandler = g_signal_connect(m_validate_webview->gobj(), "web-view-ready", G_CALLBACK(&MainWindow::webview_ready_1), this);
	m_validate_webview_loadstatus_sighandler = g_signal_connect(m_validate_webview->gobj(), "notify::load-status", G_CALLBACK(&MainWindow::loadstatus_webview_1), this);
	m_validate_webview_navpolicy_sighandler = g_signal_connect(m_validate_webview->gobj(), "navigation-policy-decision-requested", G_CALLBACK(&MainWindow::policy_decision_required_1), this);
	m_validate_webview_newwndpolicy_sighandler = g_signal_connect(m_validate_webview->gobj(), "new-window-policy-decision-requested", G_CALLBACK(&MainWindow::policy_decision_required_1), this);
	m_validate_webview_rsrcload_sighandler = g_signal_connect(m_validate_webview->gobj(), "resource-load-finished", G_CALLBACK(&MainWindow::resourceload_finished_1), this);
	m_validate_webview_consolemsg_sighandler = g_signal_connect(m_validate_webview->gobj(), "console-message", G_CALLBACK(&MainWindow::console_message_1), this);
}

void CFMUValidate::MainWindow::remove_validate_sighandlers(void)
{
        if (m_validate_webview_close_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_close_sighandler);
		m_validate_webview_close_sighandler = 0;
        }
        if (m_validate_webview_create_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_create_sighandler);
		m_validate_webview_create_sighandler = 0;
        }
	if (m_validate_webview_ready_sighandler) {
		g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_ready_sighandler);
		m_validate_webview_ready_sighandler = 0;
	}
        if (m_validate_webview_loadstatus_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_loadstatus_sighandler);
		m_validate_webview_loadstatus_sighandler = 0;
        }
        if (m_validate_webview_navpolicy_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_navpolicy_sighandler);
		m_validate_webview_navpolicy_sighandler = 0;
        }
        if (m_validate_webview_newwndpolicy_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_newwndpolicy_sighandler);
		m_validate_webview_newwndpolicy_sighandler = 0;
        }
        if (m_validate_webview_rsrcload_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_rsrcload_sighandler);
		m_validate_webview_rsrcload_sighandler = 0;
        }
        if (m_validate_webview_consolemsg_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_consolemsg_sighandler);
		m_validate_webview_consolemsg_sighandler = 0;
        }
}

void CFMUValidate::MainWindow::set_app(CFMUValidate *app)
{
	m_app = app;
}

void CFMUValidate::MainWindow::prepare_webview(void)
{
	remove_portal_sighandlers();
	delete m_portal_webview;
	m_portal_webview = Gtk::manage(Glib::wrap(webkit_web_view_new()));
	m_portal_webview->show();
	add_portal_sighandlers();
	if (m_scrolledwindowbrowser) {
		m_scrolledwindowbrowser->remove();
		m_scrolledwindowbrowser->add(*m_portal_webview);
	}
	WebKitWebView *webview = WEBKIT_WEB_VIEW(m_portal_webview->gobj());
	webkit_web_view_load_uri(webview, "https://www.public.cfmu.eurocontrol.int/PUBPORTAL/gateway/spec/index.html");
	//webkit_web_view_load_uri(webview, "http://www.baycom.org/~tom");
	set_wvstate(wvstate_loadmainpage);
}

void CFMUValidate::MainWindow::set_wvstate(wvstate_t newstate)
{
	if (newstate == m_wvstate)
		return;
	m_wvstate = newstate;
	if (m_buttonvalidate)
		m_buttonvalidate->set_sensitive(m_wvstate == wvstate_ready);
}

void CFMUValidate::MainWindow::validate_clicked(void)
{
	if (false) {
		WebKitWebView *webview(0);
		if (m_validate_webview)
			webview = WEBKIT_WEB_VIEW(m_validate_webview->gobj());
		else if (m_portal_webview)
			webview = WEBKIT_WEB_VIEW(m_portal_webview->gobj());
		else
			return;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMNodeList *elements(webkit_dom_document_get_elements_by_tag_name(document, "*"));
		gulong element_count(webkit_dom_node_list_get_length(elements));
		if (true)
			std::cerr << "Elements: " << element_count << std::endl;
		for (gulong i = 0; i < element_count; i++) {
			WebKitDOMNode *element(webkit_dom_node_list_item(elements, i));
			{
				const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
				if (typ)
					std::cerr << "element " << i << " type " << typ << std::endl;
			}
			{
				gchar *txt(webkit_dom_node_get_text_content(element));
				if (txt) {
					if (*txt)
						std::cerr << "element " << i << " text \"" << txt << "\"" << std::endl;
					g_free(txt);
				}
			}
			if (WEBKIT_DOM_IS_HTML_INPUT_ELEMENT(element)) {
				char *element_type;
				g_object_get(element, "type", &element_type, NULL);
				std::cerr << "element " << i << " type " << element_type << std::endl;
				g_free(element_type);
			}
			if (WEBKIT_DOM_IS_HTML_ELEMENT(element)) {
				char *class_name = webkit_dom_html_element_get_class_name(WEBKIT_DOM_HTML_ELEMENT(element));
				if (class_name) {
					if (*class_name)
						std::cerr << "element " << i << " html class name " << class_name << std::endl;
					g_free(class_name);
				}
				char *innertxt = webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(element));
				if (innertxt) {
					if (*innertxt)
						std::cerr << "element " << i << " html inner text " << innertxt << std::endl;
					g_free(innertxt);
				}
			}
		}
		g_object_unref(elements);
	}
	if (false) {
		WebKitWebView *webview(0);
		if (m_validate_webview)
			webview = WEBKIT_WEB_VIEW(m_validate_webview->gobj());
		else if (m_portal_webview)
			webview = WEBKIT_WEB_VIEW(m_portal_webview->gobj());
		else
			return;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMNodeList *elements(webkit_dom_document_get_elements_by_tag_name(document, "*"));
		gulong element_count(webkit_dom_node_list_get_length(elements));
		for (gulong i = 0; i < element_count; i++) {
			WebKitDOMNode *element(webkit_dom_node_list_item(elements, i));
			if (!WEBKIT_DOM_IS_HTML_ELEMENT(element))
				continue;
			gchar *txt(webkit_dom_node_get_text_content(element));
			if (!txt)
				continue;
			{
				bool ised(!strcmp(txt, "Free Text Editor"));
				g_free(txt);
				if (!ised)
					continue;
			}
			if (!webkit_dom_node_has_attributes(element))
				continue;
			{
				WebKitDOMNamedNodeMap *map(webkit_dom_node_get_attributes(element));
				std::cerr << "Attr:";
				gulong nrels(webkit_dom_named_node_map_get_length(map));
				for (gulong i = 0; i < nrels; ++i) {
					WebKitDOMNode *n(webkit_dom_named_node_map_item(map, i));
					if (!n)
						continue;
					gchar *lname(webkit_dom_node_get_local_name(n));
					if (lname) {
						if (*lname)
							std::cerr << ' ' << lname << ':';
						g_free(lname);
					}
					gchar *txt(webkit_dom_node_get_node_value(n));
					if (!txt)
						continue;
					if (*txt)
						std::cerr << ' ' << txt;
					g_free(txt);
				}
				std::cerr << std::endl;

				//WebKitDOMNode *id(webkit_dom_named_node_map_get_named_item(map, "textContent"));
				//std::cerr << "onclick: " << webkit_dom_node_get_node_value(id) << std::endl;
			}
			std::cerr << "Free Text Editor Element found!" << std::endl;
			{
				gchar *uri(webkit_dom_node_get_base_uri(element));
				if (uri) {
					if (*uri)
						std::cerr << "  Base URI: " << uri << std::endl;
					g_free(uri);
				}
			}
		}
	}
	if (!m_validate_webview)
		return;
	{
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview->gobj()));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		if (!document)
			return;
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD"));
		if (!element || !WEBKIT_DOM_IS_HTML_TEXT_AREA_ELEMENT(element))
			return;
		if (!m_textviewfplan)
			return;
		Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewfplan->get_buffer());
		if (!buf)
			return;
		const Glib::ustring& txt(buf->get_text());
		webkit_dom_html_text_area_element_set_value(WEBKIT_DOM_HTML_TEXT_AREA_ELEMENT(element), txt.c_str());
		if (txt.empty())
			return;
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL");
		if (!element)
			return;
		zap_errors();
		webkit_web_view_execute_script(webview, "document.getElementById(\"FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL\").click();");
		set_wvstate(wvstate_query);
	}
}
//<button type="button" class="portal_buttonToolsButton_enabled" id="FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL">Validate</button>
//<textarea class="gwt-TextArea ifpuv_freeTextPanel_fplTextArea" id="FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD"></textarea>
//<table cellspacing="0" cellpadding="0" class="cfmu_DisclosurePanel cfmu_DisclosurePanel-closed" id="FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA"><tbody><tr><td align="left" style="vertical-align: top; "><a href="javascript:void(0);" style="display: block; " class="header"><table cellspacing="0" cellpadding="0" class="portal_fullWidthStyle"><tbody><tr><td align="left" style="vertical-align: top; " width="" height="" rowspan="1"><div class="cfmu_DisclosurePanel_Label cfmu_DisclosurePanel_Enabled">Validation&nbsp;Results</div></td><td align="left" style="vertical-align: top; width: 100%; " width="" height=""><div class="gwt-HTML">&nbsp;</div></td><td align="left" style="vertical-align: top; " width="" height="" rowspan="1"><div class="portal_disclosurePanelRightTextStyle_0"></div></td></tr></tbody></table></a></td></tr><tr><td align="left" style="vertical-align: top; "><div style="overflow-x: hidden; overflow-y: hidden; display: none; padding-top: 0px; padding-right: 0px; padding-bottom: 0px; padding-left: 12px; "><table cellspacing="0" cellpadding="0" class="portal_cfmuDisclosurePanel_contentContainer content"><tbody><tr><td align="left" style="vertical-align: top; "><table cellspacing="0" cellpadding="0" class="portal_fullSizeStyle portal_fullWidthStyle" style="display: none; "><tbody><tr><td align="left" style="vertical-align: top; "><div class="portal_warningLabel" style="display: none; "></div></td></tr><tr><td align="left" style="vertical-align: top; "><div class="portal_dataValue" style="text-align: left; display: none; ">NO ERRORS</div></td></tr><tr><td align="left" style="vertical-align: top; "><div style="padding-top: 0px; padding-right: 0px; padding-bottom: 0px; padding-left: 0px; position: relative; display: none; " id="FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA.ERRORS_DATA_TABLE" class="portal_fullWidthStyle"><div tabindex="0" style="display: none; position: relative; " class="portal_sliderBar"><input type="text" tabindex="-1" style="opacity: 0; height: 1px; width: 1px; z-index: -1; overflow-x: hidden; overflow-y: hidden; position: absolute; "><div style="position: absolute; left: 0px; " class="portal_sliderBarLine"></div><div class="portal_sliderBarKnob" style="position: absolute; left: -1px; "></div><div style="position: absolute; left: -1px; visibility: visible; " class="portal_sliderBarTick"></div><div style="position: absolute; left: -1px; visibility: visible; " class="portal_sliderBarLabel">1</div><div style="position: absolute; left: 0px; visibility: visible; " class="portal_sliderBarTick"></div><div style="position: absolute; left: 0px; visibility: visible; " class="portal_sliderBarLabel">1</div></div><div></div><table class="portal_noScreen portal_autopagetablecontainer_headerbackground portal_noPrint" cellpadding="0" cellspacing="0"><colgroup><col></colgroup><tbody><tr class="portal_autoPageTableContainer_headerRowStyle"><td class="portal_autoPageTableContainer_headerCellStyle"><div class="gwt-HTML">Errors</div></td></tr></tbody></table><div style="width: 100%; padding-top: 0px; padding-right: 0px; padding-bottom: 0px; padding-left: 0px; margin-top: 0px; margin-right: 0px; margin-bottom: 0px; margin-left: 0px; border-top-style: none; border-right-style: none; border-bottom-style: none; border-left-style: none; border-width: initial; border-color: initial; border-image: initial; position: relative; "><table cellpadding="0" cellspacing="0" class="portal_decotable"><colgroup><col></colgroup><tbody></tbody></table></div></div></td></tr><tr><td align="left" style="vertical-align: top; "><table cellspacing="0" cellpadding="0" class="portal_fullWidthStyle"><tbody><tr><td align="left" style="vertical-align: top; "><div class="portal_dataValue" style="display: none; ">No routes found.</div></td></tr><tr><td align="left" style="vertical-align: top; "><table cellspacing="0" cellpadding="0" id="FREE_TEXT_EDITOR.PROPOSE_ROUTES_RESULTS_AREA.ORIGINAL_ROUTE_DATA_TABLE"><tbody><tr><td align="left" style="vertical-align: top; "><table class="portal_headerTableDecorator_headerbackground" cellpadding="0" cellspacing="0"><colgroup><col></colgroup><tbody><tr class="portal_headerTableDecorator_headerRowStyle"></tr></tbody></table></td></tr><tr><td align="left" style="vertical-align: top; "><table cellpadding="0" cellspacing="0" class="portal_decotable"><colgroup><col></colgroup><tbody></tbody></table></td></tr></tbody></table></td></tr></tbody></table></td></tr></tbody></table></td></tr></tbody></table></div></td></tr></tbody></table>
//<div class="portal_dataValue" style="text-align: left; ">NO ERRORS</div>
//<td class="portal_decotableCell ">ROUTE126: FLIGHT RULES Y WITH NO VFR PART.</td>
//<td class="portal_decotableCell ">ROUTE130: UNKNOWN DESIGNATOR HOC3Z</td>

void CFMUValidate::MainWindow::quit_clicked(void)
{
	if (m_app)
		m_app->quit();
}

void CFMUValidate::MainWindow::about_clicked(void)
{
	if (m_aboutwindow)
		m_aboutwindow->show();
}

void CFMUValidate::MainWindow::about_done(int resp)
{
	if (m_aboutwindow)
		m_aboutwindow->hide();
}

void CFMUValidate::MainWindow::fpl_changed(void)
{
	if (!m_textviewresults)
		return;
	Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewresults->get_buffer());
	if (!buf)
		return;
	buf->set_text("");
}

void CFMUValidate::MainWindow::parse_results(void)
{
	if (!m_validate_webview)
		return;
	typedef std::pair<guint,bool> lvlvisible_t;
	typedef std::pair<WebKitDOMNode *,lvlvisible_t> nodelvlvisible_t;
	std::stack<nodelvlvisible_t> subtree;
	{
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview->gobj()));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA"));
		if (!element || !WEBKIT_DOM_IS_NODE(element))
			return;
		subtree.push(std::make_pair(WEBKIT_DOM_NODE(element), std::make_pair(0U, true)));
	}
	std::ostringstream results;
	while (!subtree.empty()) {
		WebKitDOMNode *node;
		guint level;
		bool visible;
		{
			nodelvlvisible_t nl(subtree.top());
			node = nl.first;
			level = nl.second.first;
			visible = nl.second.second;
		}
		subtree.pop();
		if (WEBKIT_DOM_IS_HTML_ELEMENT(node)) {
			WebKitDOMCSSStyleDeclaration *style(webkit_dom_element_get_style(WEBKIT_DOM_ELEMENT(node)));
			if (style) {
				gchar *disp(webkit_dom_css_style_declaration_get_property_value(style, "display"));
				if (disp) {
					visible = !!strcmp(disp, "none");
					g_free(disp);
				}
			}
		}
		if (visible) {
			WebKitDOMNodeList *list(webkit_dom_node_get_child_nodes(node));
			gulong i(webkit_dom_node_list_get_length(list));
			while (i) {
				--i;
				WebKitDOMNode *ptr(webkit_dom_node_list_item(list, i));
				if (!ptr)
					continue;
				subtree.push(std::make_pair(ptr, std::make_pair(level + 1U, visible)));
			}
		}
		if (!WEBKIT_DOM_IS_HTML_ELEMENT(node))
			continue;
		char *class_name(webkit_dom_html_element_get_class_name(WEBKIT_DOM_HTML_ELEMENT(node)));
		if (false) {
			std::cerr << std::string(2U * level, ' ') << "element:";
			gchar *lname(webkit_dom_node_get_local_name(WEBKIT_DOM_NODE(node)));
			if (lname) {
				if (*lname)
					std::cerr << " name \"" << lname << "\"";
				g_free(lname);
			}
			lname = webkit_dom_html_element_get_id(WEBKIT_DOM_HTML_ELEMENT(node));
			if (lname) {
				if (*lname)
					std::cerr << " id \"" << lname << "\"";
				g_free(lname);
			}
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(node)));
			if (typ && *typ)
				std::cerr << " type \"" << typ << "\"";
			if (class_name && *class_name)
				std::cerr << " class \"" << class_name << "\"";
			WebKitDOMCSSStyleDeclaration *style(webkit_dom_element_get_style(WEBKIT_DOM_ELEMENT(node)));
			if (style) {
				gchar *disp(webkit_dom_css_style_declaration_get_property_value(style, "display"));
				if (disp) {
					if (*disp)
						std::cerr << " display \"" << disp << "\"" << std::endl;
					g_free(disp);
				}
			}
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt)
					std::cerr << " text \"" << txt << "\"";
				g_free(txt);
			}
			std::cerr << std::endl;
		}
		if (!class_name)
			continue;
		if (!visible) {
			g_free(class_name);
			continue;
		}
		if (!strcmp(class_name, "portal_dataValue") && WEBKIT_DOM_IS_HTML_DIV_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results << txt << std::endl;
					if (true)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		if (false && !strcmp(class_name, "portal_decotableCell ") && WEBKIT_DOM_IS_HTML_TABLE_CELL_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results << txt << std::endl;
					if (true)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		if (true && (!strcmp(class_name, "portal_rowOdd") || !strcmp(class_name, "portal_rowEven")) &&
		    WEBKIT_DOM_IS_HTML_TABLE_ROW_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results << txt << std::endl;
					if (true)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		g_free(class_name);
	}
	if (!m_textviewresults)
		return;
	Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewresults->get_buffer());
	if (!buf)
		return;
	buf->set_text(results.str());
	if (!results.str().empty())
		set_wvstate(wvstate_ready);
}

void CFMUValidate::MainWindow::zap_errors(void)
{
	if (!m_validate_webview)
		return;
	std::stack<WebKitDOMNode *> subtree;
	{
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview->gobj()));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA"));
		if (!element || !WEBKIT_DOM_IS_NODE(element))
			return;
		subtree.push(WEBKIT_DOM_NODE(element));
	}
	std::ostringstream results;
	while (!subtree.empty()) {
		WebKitDOMNode *node(subtree.top());
		subtree.pop();
		if (WEBKIT_DOM_IS_HTML_ELEMENT(node)) {
			char *class_name(webkit_dom_html_element_get_class_name(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (class_name) {
		     		if ((!strcmp(class_name, "portal_rowOdd") || !strcmp(class_name, "portal_rowEven")) &&
				    WEBKIT_DOM_IS_HTML_TABLE_ROW_ELEMENT(node)) {
					g_free(class_name);
					while (webkit_dom_node_has_child_nodes(node)) {
						GError *error(0);
						webkit_dom_node_remove_child(node, webkit_dom_node_get_first_child(node), &error);
						if (error) {
							std::cerr << "Cannot remove child: " << error->message << std::endl;
							g_error_free(error);
						}
					}
					continue;
				}
				g_free(class_name);
			}
		}
		{
			WebKitDOMNodeList *list(webkit_dom_node_get_child_nodes(node));
			gulong i(webkit_dom_node_list_get_length(list));
			while (i) {
				--i;
				WebKitDOMNode *ptr(webkit_dom_node_list_item(list, i));
				if (!ptr)
					continue;
				subtree.push(ptr);
			}
		}
	}
}

gboolean CFMUValidate::MainWindow::close_webview_1(WebKitWebView* webview, gpointer userdata)
{
	return static_cast<MainWindow *>(userdata)->close_webview(webview);
}

gboolean CFMUValidate::MainWindow::close_webview(WebKitWebView* webview)
{
        set_wvstate(wvstate_failed);
	hide();
	return true;
}

WebKitWebView *CFMUValidate::MainWindow::create_webview_1(WebKitWebView *webview, WebKitWebFrame *frame, gpointer userdata)
{
	return static_cast<MainWindow *>(userdata)->create_webview(webview, frame);
}

WebKitWebView *CFMUValidate::MainWindow::create_webview(WebKitWebView *webview, WebKitWebFrame *frame)
{
	std::cerr << "new webview!" << std::endl;
	if (m_validate_webview && webview == WEBKIT_WEB_VIEW(m_validate_webview->gobj()))
		return 0;
        g_assert(m_portal_webview && webview == WEBKIT_WEB_VIEW(m_portal_webview->gobj()));
	if (m_wvstate != wvstate_loadmainpage)
		return 0;
	remove_validate_sighandlers();
	delete m_validate_webview;
	m_validate_webview = Gtk::manage(Glib::wrap(webkit_web_view_new()));
	webkit_web_view_set_settings(WEBKIT_WEB_VIEW(m_validate_webview->gobj()), webkit_web_view_get_settings(WEBKIT_WEB_VIEW(m_portal_webview->gobj())));
	add_validate_sighandlers();
	set_wvstate(wvstate_loadquerypage);
	WebKitWebView *wv(WEBKIT_WEB_VIEW(m_validate_webview->gobj()));
	return wv;
}

gboolean CFMUValidate::MainWindow::webview_ready_1(WebKitWebView* webview, gpointer userdata)
{
	return static_cast<MainWindow *>(userdata)->webview_ready(webview);
}

gboolean CFMUValidate::MainWindow::webview_ready(WebKitWebView* webview)
{
	if (true)
		std::cerr << "WebView: ready" << std::endl;
	if (m_validate_webview && webview == WEBKIT_WEB_VIEW(m_validate_webview->gobj())) {
		if (m_wvstate == wvstate_loadquerypage && m_scrolledwindowbrowser) {
			m_scrolledwindowbrowser->remove();
			m_scrolledwindowbrowser->add(*m_validate_webview);
		}
		m_validate_webview->show_all();
		return false;
	}
	g_assert(m_portal_webview && webview == WEBKIT_WEB_VIEW(m_portal_webview->gobj()));
	m_portal_webview->show_all();
	return false;
}

void CFMUValidate::MainWindow::loadstatus_webview_1(WebKitWebView *webview, GParamSpec *pspec, gpointer userdata)
{
	static_cast<MainWindow *>(userdata)->loadstatus_webview(webview, pspec);
}

void CFMUValidate::MainWindow::loadstatus_webview(WebKitWebView *webview, GParamSpec *pspec)
{
        WebKitLoadStatus status(webkit_web_view_get_load_status(webview));
        GObject *object(G_OBJECT(webview));
        g_assert((m_portal_webview && webview == WEBKIT_WEB_VIEW(m_portal_webview->gobj())) ||
		 (m_validate_webview && webview == WEBKIT_WEB_VIEW(m_validate_webview->gobj())));
        g_object_freeze_notify(object);
	switch (status) {
        case WEBKIT_LOAD_FINISHED:
		if (true)
			std::cerr << "WebView: load finished" << std::endl;
		if (m_validate_webview && webview == WEBKIT_WEB_VIEW(m_validate_webview->gobj()))
			m_validate_webview->show_all();
		else if (m_portal_webview && webview == WEBKIT_WEB_VIEW(m_portal_webview->gobj()))
			m_portal_webview->show_all();
		break;

        case WEBKIT_LOAD_PROVISIONAL:
		if (true)
			std::cerr << "WebView: load provisional" << std::endl;
                break;

        case WEBKIT_LOAD_COMMITTED:
		if (true)
			std::cerr << "WebView: load committed" << std::endl;
                break;

        case WEBKIT_LOAD_FIRST_VISUALLY_NON_EMPTY_LAYOUT:
		if (true)
			std::cerr << "WebView: load first visually nonempty layout" << std::endl;
                break;

        case WEBKIT_LOAD_FAILED:
		if (true)
			std::cerr << "WebView: load failed" << std::endl;
		set_wvstate(wvstate_failed);
                break;

        default:
                break;
        }
        g_object_thaw_notify(object);
}

gboolean CFMUValidate::MainWindow::policy_decision_required_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
							      WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision, gpointer userdata)
{
	return static_cast<MainWindow *>(userdata)->policy_decision_required(webview, webframe, request, action, decision);
}

gboolean CFMUValidate::MainWindow::policy_decision_required(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
							    WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision)
{
        g_assert((m_portal_webview && webview == WEBKIT_WEB_VIEW(m_portal_webview->gobj())) ||
		 (m_validate_webview && webview == WEBKIT_WEB_VIEW(m_validate_webview->gobj())));
	WebKitWebNavigationReason reason(webkit_web_navigation_action_get_reason(action));
        gint button(webkit_web_navigation_action_get_button(action));
        gint state(webkit_web_navigation_action_get_modifier_state(action));
        const char *uri(webkit_network_request_get_uri(request));
	if (true)
		std::cerr << "policy_decision_required: reason " << reason << " button " << button << " state " << state
			  << " uri " << uri << std::endl;
	return false;
}

void CFMUValidate::MainWindow::resourceload_finished_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource, gpointer userdata)
{
	static_cast<MainWindow *>(userdata)->resourceload_finished(webview, webframe, webresource);
}

void CFMUValidate::MainWindow::resourceload_finished(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource)
{
        g_assert((m_portal_webview && webview == WEBKIT_WEB_VIEW(m_portal_webview->gobj())) ||
		 (m_validate_webview && webview == WEBKIT_WEB_VIEW(m_validate_webview->gobj())));
 	if (true) {
		time_t t(time(0));
		char buf[64];
		std::cerr << "resource load finished! " << ctime_r(&t, buf) << " state " << (int)m_wvstate << std::endl;
	}
	switch (m_wvstate) {
	case wvstate_loadmainpage:
	{
		if (!m_portal_webview || webview != WEBKIT_WEB_VIEW(m_portal_webview->gobj()))
			break;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "IFPUV_LAUNCH_AREA.FREE_TEXT_EDIT_LINK_LABEL"));
		if (!element)
			break;
		if (false)
			std::cerr << "Button found! Injecting script" << std::endl;
		webkit_web_view_execute_script(webview, "document.getElementById(\"IFPUV_LAUNCH_AREA.FREE_TEXT_EDIT_LINK_LABEL\").click();");
		set_wvstate(wvstate_loadquerypage);
		break;
	}

	case wvstate_loadquerypage:
	{
		if (!m_validate_webview || webview != WEBKIT_WEB_VIEW(m_validate_webview->gobj()))
			break;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL"));
		if (!element)
			break;
		if (true) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Validate element type " << typ << std::endl;
		}
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD");
		if (!element)
			break;
		if (true) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "FPL entry element type " << typ << std::endl;
		}
		if (!WEBKIT_DOM_IS_HTML_TEXT_AREA_ELEMENT(element))
			break;
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA.ERRORS_DATA_TABLE");
		if (!element)
			break;
		if (true) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Result Errors element type " << typ << std::endl;
		}
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA");
		if (!element)
			break;
		webkit_dom_event_target_add_event_listener(WEBKIT_DOM_EVENT_TARGET(element), "DOMSubtreeModified", G_CALLBACK(&MainWindow::result_change_1), true, this);
		if (true) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Results element type " << typ << std::endl;
		}
		set_wvstate(wvstate_ready);
		break;
	}

	case wvstate_query:
	{
		if (!m_validate_webview || webview != WEBKIT_WEB_VIEW(m_validate_webview->gobj()))
			break;
		set_wvstate(wvstate_waitresult);
		parse_results();
		break;
	}

	default:
		break;
	}
}

gboolean CFMUValidate::MainWindow::result_change_1(WebKitDOMElement *target, WebKitDOMEvent *e, gpointer userdata)
{
	return static_cast<MainWindow *>(userdata)->result_change(target, e);
}

gboolean CFMUValidate::MainWindow::result_change(WebKitDOMElement *target, WebKitDOMEvent *e)
{
	std::cerr << "result change! state " << (int)m_wvstate << std::endl;
	switch (m_wvstate) {
	case wvstate_waitresult:
		parse_results();
		break;

	default:
		break;
	}
	return true;
}

gboolean CFMUValidate::MainWindow::console_message_1(WebKitWebView *webview, gchar *message, gint line, gchar *source_id, gpointer userdata)
{
	return static_cast<MainWindow *>(userdata)->console_message(webview, message, line, source_id);
}

gboolean CFMUValidate::MainWindow::console_message(WebKitWebView *webview, gchar *message, gint line, gchar *source_id)
{
	if (true) {
		std::cerr << message << " (" << (source_id ? "*null*" : source_id) << ':' << line << ')' << std::endl;
	}
	return true;
}

CFMUValidate::CFMUValidate(int& argc, char**& argv, bool set_locale)
	: Gtk::Main(argc, argv, set_locale), m_mainwindow(0)
{
	m_builder = load_glade_file("cfmuvalidate" UIEXT);
        m_builder->get_widget_derived("cfmuvalidatewindow", m_mainwindow);
	if (m_mainwindow)
		m_mainwindow->set_app(this);
	{
		guint nsigids(0);
		guint *sigids(g_signal_list_ids(WEBKIT_TYPE_DOM_HTML_ELEMENT, &nsigids));
		std::cerr << "DOM HTML Element signal list: " << nsigids << std::endl;
		for (guint i(0); i < nsigids; ++i) {
			GSignalQuery q;
			g_signal_query(sigids[i], &q);
			std::cerr << "  " << q.signal_name << std::endl;
		}
	}
}

CFMUValidate::~CFMUValidate()
{
	delete m_mainwindow;
}

void CFMUValidate::run(void)
{
	if (m_mainwindow)
		Gtk::Main::run(*m_mainwindow);
}

int main(int argc, char *argv[])
{
        try {
                Glib::init();
                Glib::set_application_name("CFMU Validate");
                Glib::thread_init();
                CFMUValidate app(argc, argv);
                app.run();
                return 0;
        } catch(const Glib::FileError& ex) {
                std::cerr << "FileError: " << ex.what() << std::endl;
                return 1;
        } catch(const Gtk::BuilderError& ex) {
                std::cerr << "BuilderError: " << ex.what() << std::endl;
                return 1;
        }
        return 1;
}
