//
// C++ Implementation: FlightDeckWindow
//
// Description: Flight Deck Windows, Altitude Dialog
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2011, 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"
#include <iomanip>

#include "flightdeckwindow.h"

void FlightDeckWindow::map_dialog_open(void)
{
	if (!m_mapdialog)
		return;
	m_mapdialog->set_flags(m_navarea.map_get_drawflags());
	m_mapdialog->set_mapup(m_navarea.map_get_mapup());
	m_mapdialog->show();
}

void FlightDeckWindow::map_settings_update(void)
{
	if (!m_mapdialog)
		return;
	m_navarea.map_set_drawflags(m_mapdialog->get_flags());
	m_navarea.map_set_mapup(m_mapdialog->get_mapup());
	m_navarea.map_set_declutter(NavArea::dcltr_normal);
	if (m_wptdrawingarea)
		*m_wptdrawingarea = m_navarea.map_get_declutter_drawflags();
	if (m_fpldrawingarea)
		*m_fpldrawingarea = m_navarea.map_get_declutter_drawflags();
	if (m_menuid <= 300)
		set_menu_button(5, m_navarea.map_get_declutter_name());
}

const unsigned int FlightDeckWindow::MapDialog::nrcheckbuttons;
const struct FlightDeckWindow::MapDialog::checkbutton FlightDeckWindow::MapDialog::checkbuttons[nrcheckbuttons] = {
	{ "mapcheckbuttontopo", MapRenderer::drawflags_topo },
	{ "mapcheckbuttonnorth", MapRenderer::drawflags_northpointer },
	{ "mapcheckbuttonterrain", MapRenderer::drawflags_terrain },
	{ "mapcheckbuttonterrainnames", MapRenderer::drawflags_terrain_names },
	{ "mapcheckbuttonterrainborders", MapRenderer::drawflags_terrain_borders },
	{ "mapcheckbuttonroute", MapRenderer::drawflags_route },
	{ "mapcheckbuttonroutelabels", MapRenderer::drawflags_route_labels },
	{ "mapcheckbuttontrack", MapRenderer::drawflags_track },
	{ "mapcheckbuttontrackpoints", MapRenderer::drawflags_track_points },
	{ "mapcheckbuttonnavaidvor", MapRenderer::drawflags_navaids_vor },
	{ "mapcheckbuttonnavaiddme", MapRenderer::drawflags_navaids_dme },
	{ "mapcheckbuttonnavaidndb", MapRenderer::drawflags_navaids_ndb },
	{ "mapcheckbuttonwaypointlow", MapRenderer::drawflags_waypoints_low },
	{ "mapcheckbuttonwaypointhigh", MapRenderer::drawflags_waypoints_high },
	{ "mapcheckbuttonwaypointrnav", MapRenderer::drawflags_waypoints_rnav },
	{ "mapcheckbuttonwaypointterminal", MapRenderer::drawflags_waypoints_terminal },
	{ "mapcheckbuttonwaypointvfr", MapRenderer::drawflags_waypoints_vfr },
	{ "mapcheckbuttonwaypointuser", MapRenderer::drawflags_waypoints_user },
	{ "mapcheckbuttonairportciv", MapRenderer::drawflags_airports_civ },
	{ "mapcheckbuttonairportmil", MapRenderer::drawflags_airports_mil },
	{ "mapcheckbuttonairportloccone", MapRenderer::drawflags_airports_localizercone },
	{ "mapcheckbuttonairspaceab", MapRenderer::drawflags_airspaces_ab },
	{ "mapcheckbuttonairspacecd", MapRenderer::drawflags_airspaces_cd },
	{ "mapcheckbuttonairspaceefg", MapRenderer::drawflags_airspaces_efg },
	{ "mapcheckbuttonairspacespecialuse", MapRenderer::drawflags_airspaces_specialuse },
	{ "mapcheckbuttonairspacefillground", MapRenderer::drawflags_airspaces_fill_ground },
	{ "mapcheckbuttonairwaylow", MapRenderer::drawflags_airways_low },
	{ "mapcheckbuttonairwayhigh", MapRenderer::drawflags_airways_high },
};

FlightDeckWindow::MapDialog::MapupModelColumns::MapupModelColumns(void)
{
	add(m_text);
}

FlightDeckWindow::MapDialog::MapDialog(BaseObjectType *castitem, const Glib::RefPtr<Builder>& refxml)
	: Gtk::Window(castitem), m_mapcomboboxup(0), m_buttonrevert(0), m_buttondone(0),
	  m_curflags(MapRenderer::drawflags_all), m_flags(MapRenderer::drawflags_all),
	  m_curmapup(NavArea::mapup_north), m_mapup(NavArea::mapup_north)
{
#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        signal_delete_event().connect(sigc::mem_fun(*this, &MapDialog::on_delete_event));
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
	refxml->get_widget("mapcomboboxup", m_mapcomboboxup);
	refxml->get_widget("mapbuttonrevert", m_buttonrevert);
	refxml->get_widget("mapbuttondone", m_buttondone);
	m_mapupstore = Gtk::ListStore::create(m_mapupcolumns);
	{
                Gtk::TreeModel::Row row(*(m_mapupstore->append()));
                row[m_mapupcolumns.m_text] = "North Up";
                row = *(m_mapupstore->append());
                row[m_mapupcolumns.m_text] = "Track Up";
                row = *(m_mapupstore->append());
                row[m_mapupcolumns.m_text] = "Heading Up";
	}
        if (m_mapcomboboxup) {
                m_mapcomboboxup->clear();
                m_mapcomboboxup->set_model(m_mapupstore);
                m_mapcomboboxup->set_entry_text_column(m_mapupcolumns.m_text);
                m_mapcomboboxup->pack_start(m_mapupcolumns.m_text);
		m_connmapup = m_mapcomboboxup->signal_changed().connect(sigc::mem_fun(*this, &MapDialog::mapup_changed));
        }
	if (m_buttonrevert)
		m_connrevert = m_buttonrevert->signal_clicked().connect(sigc::mem_fun(*this, &MapDialog::revert_clicked));
	if (m_buttondone)
		m_conndone = m_buttondone->signal_clicked().connect(sigc::mem_fun(*this, &MapDialog::done_clicked));
	for (unsigned int i = 0; i < nrcheckbuttons; ++i) {
		m_checkbutton[i] = 0;
		refxml->get_widget(checkbuttons[i].widgetname, m_checkbutton[i]);
		if (!m_checkbutton[i])
			continue;
		m_conncheckbutton[i] = m_checkbutton[i]->signal_toggled().connect(sigc::bind(sigc::mem_fun(*this, &MapDialog::flag_toggled), i));
	}
	set_flags(MapRenderer::drawflags_all);
}

FlightDeckWindow::MapDialog::~MapDialog()
{
}

void FlightDeckWindow::MapDialog::set_flags(MapRenderer::DrawFlags flags)
{
	m_curflags = m_flags = flags;
	for (unsigned int i = 0; i < nrcheckbuttons; ++i)
		m_conncheckbutton[i].block();
	for (unsigned int i = 0; i < nrcheckbuttons; ++i)
		if (m_checkbutton[i])
			m_checkbutton[i]->set_active(!!(m_flags & checkbuttons[i].drawflag));
	for (unsigned int i = 0; i < nrcheckbuttons; ++i)
		m_conncheckbutton[i].unblock();
}

void FlightDeckWindow::MapDialog::set_mapup(NavArea::mapup_t up)
{
	m_curmapup = m_mapup = up;
	m_connmapup.block();
	if (m_mapcomboboxup)
		m_mapcomboboxup->set_active((int)m_mapup);
	m_connmapup.unblock();
}

bool FlightDeckWindow::MapDialog::on_delete_event(GdkEventAny* event)
{
        hide();
	m_signal_done.emit();
        return true;
}

void FlightDeckWindow::MapDialog::done_clicked(void)
{
        hide();
	m_signal_done.emit();
}

void FlightDeckWindow::MapDialog::revert_clicked(void)
{
	set_flags(m_curflags);
	set_mapup(m_curmapup);
}

void FlightDeckWindow::MapDialog::flag_toggled(unsigned int nr)
{
	if (nr >= nrcheckbuttons || !m_checkbutton[nr])
		return;
	if (m_checkbutton[nr]->get_active())
		m_flags |= checkbuttons[nr].drawflag;
	else
		m_flags &= ~checkbuttons[nr].drawflag;
}

void FlightDeckWindow::MapDialog::mapup_changed(void)
{
	if (!m_mapcomboboxup)
		return;
	int x(m_mapcomboboxup->get_active_row_number());
	if (x == -1)
		return;
	m_mapup = (NavArea::mapup_t)x;
}
