//
// C++ Implementation: FlightDeckWindow
//
// Description: Flight Deck Windows, Gyro Instrument Drawing Areas
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2011, 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"

#include "flightdeckwindow.h"

#include <memory>
#include <iomanip>

FlightDeckWindow::RouteProfileDrawingArea::RouteProfileDrawingArea(BaseObjectType *castitem, const Glib::RefPtr<Builder>& refxml)
	: Gtk::DrawingArea(castitem), m_engine(0), m_route(0), m_scaledist(1), m_scaleelev(0.1), m_origindist(0), m_originelev(0),
	  m_dragx(-1), m_dragy(-1)

{
#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
#ifdef HAVE_GTKMM2
	signal_expose_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_expose_event));
#endif
#ifdef HAVE_GTKMM3
        signal_draw().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_draw));
#endif
        signal_show().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_show));
        signal_hide().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_hide));
        signal_button_press_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_button_press_event));
        signal_button_release_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_button_release_event));
        signal_motion_notify_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_motion_notify_event));
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
	m_dispatch.connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::async_result));
	//set_size_request(128, 128);
	//add_events(Gdk::EXPOSURE_MASK | Gdk::PROPERTY_CHANGE_MASK | Gdk::POINTER_MOTION_MASK | Gdk::POINTER_MOTION_HINT_MASK);
}

FlightDeckWindow::RouteProfileDrawingArea::RouteProfileDrawingArea(void)
	: m_engine(0), m_route(0), m_scaledist(1), m_scaleelev(0.1), m_origindist(0), m_originelev(0),
	  m_dragx(-1), m_dragy(-1)
{
#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
#ifdef HAVE_GTKMM2
	signal_expose_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_expose_event));
#endif
#ifdef HAVE_GTKMM3
        signal_draw().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_draw));
#endif
        signal_show().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_show));
        signal_hide().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_hide));
        signal_button_press_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_button_press_event));
        signal_button_release_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_button_release_event));
        signal_motion_notify_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_motion_notify_event));
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
	m_dispatch.connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::async_result));
	//set_size_request(128, 128);
	//add_events(Gdk::EXPOSURE_MASK | Gdk::PROPERTY_CHANGE_MASK | Gdk::POINTER_MOTION_MASK | Gdk::POINTER_MOTION_HINT_MASK);
}

FlightDeckWindow::RouteProfileDrawingArea::~RouteProfileDrawingArea()
{
	async_cancel();
}

void FlightDeckWindow::RouteProfileDrawingArea::set_engine(Engine& eng)
{
	async_cancel();
	m_engine = &eng;
	m_profile.clear();
	route_change();
	
}

void FlightDeckWindow::RouteProfileDrawingArea::set_route(const FPlanRoute& route)
{
	async_cancel();
	m_route = &route;
	m_profile.clear();
	route_change();
}

void FlightDeckWindow::RouteProfileDrawingArea::route_change(void)
{
	async_cancel();
	m_profile.clear();
	if (!get_is_drawable())
		return;
	async_startquery();
}

void FlightDeckWindow::RouteProfileDrawingArea::async_startquery(void)
{
	if (!m_engine || !m_route || m_route->get_nrwpt() < 2)
		return;
	async_cancel();
	async_clear();
	m_profilequery = m_engine->async_elevation_profile((*m_route)[0].get_coord(), (*m_route)[1].get_coord(), 5.);
	async_connectdone();
}

void FlightDeckWindow::RouteProfileDrawingArea::dist_zoom_in(void)
{
	set_dist_zoom(get_dist_zoom() * 2.);
}

void FlightDeckWindow::RouteProfileDrawingArea::dist_zoom_out(void)
{
	set_dist_zoom(get_dist_zoom() * 0.5);
}

double FlightDeckWindow::RouteProfileDrawingArea::get_dist_zoom(void) const
{
	return m_scaledist;
}

void FlightDeckWindow::RouteProfileDrawingArea::set_dist_zoom(double z)
{
	double zz(std::max(std::min(z, 100.), 0.01));
	if (zz == m_scaledist)
		return;
	m_scaledist = zz;
	if (get_is_drawable())
		queue_draw();
}

void FlightDeckWindow::RouteProfileDrawingArea::elev_zoom_in(void)
{
	set_elev_zoom(get_elev_zoom() * 2.);
}

void FlightDeckWindow::RouteProfileDrawingArea::elev_zoom_out(void)
{
	set_elev_zoom(get_elev_zoom() * 0.5);
}

double FlightDeckWindow::RouteProfileDrawingArea::get_elev_zoom(void) const
{
	return m_scaleelev;
}

void FlightDeckWindow::RouteProfileDrawingArea::set_elev_zoom(double z)
{
	double zz(std::max(std::min(z, 10.), 0.001));
	if (zz == m_scaleelev)
		return;
	m_scaleelev = zz;
	if (get_is_drawable())
		queue_draw();
}

#ifdef HAVE_GTKMM2
bool FlightDeckWindow::RouteProfileDrawingArea::on_expose_event(GdkEventExpose *event)
{
        Glib::RefPtr<Gdk::Window> wnd(get_window());
        if (!wnd)
                return false;
        Cairo::RefPtr<Cairo::Context> cr(wnd->create_cairo_context());
	if (event) {
		GdkRectangle *rects;
		int n_rects;
		gdk_region_get_rectangles(event->region, &rects, &n_rects);
		for (int i = 0; i < n_rects; i++)
			cr->rectangle(rects[i].x, rects[i].y, rects[i].width, rects[i].height);
		cr->clip();
		g_free(rects);
	}
        return on_draw(cr);
}
#endif

bool FlightDeckWindow::RouteProfileDrawingArea::on_draw(const Cairo::RefPtr<Cairo::Context>& cr)
{
	cr->save();
	// background
	cr->set_source_rgb(1.0, 1.0, 1.0);
	cr->paint();
	// scaling
	Gtk::Allocation allocation = get_allocation();
	const int width = allocation.get_width();
	const int height = allocation.get_height();
        cr->select_font_face("sans", Cairo::FONT_SLANT_NORMAL, Cairo::FONT_WEIGHT_NORMAL);
        cr->set_font_size(12);
	int alt_width;
	int dist_height;
	int dist_line_height;
	double elevinc, elevstart;
	double distinc, diststart;
	{
		Cairo::TextExtents ext;
		cr->get_text_extents("99999", ext);
		alt_width = ext.width + 4;
		dist_line_height = ext.height + 4;
		dist_height = 2 * dist_line_height + 5;
		elevinc = pow(10., std::max(0., ceil(log10(ext.height * 5 / m_scaleelev))));
		elevstart = elevinc * ceil(m_originelev / elevinc);
		cr->get_text_extents("999", ext);
		distinc = pow(10., std::max(0., ceil(log10(ext.width * 3 / m_scaledist))));
		diststart = distinc * ceil(m_origindist / distinc);
	}
	// draw location labels
	if (m_route) {
		set_color_fplan(cr);
		for (unsigned int i = 0, j = m_route->get_nrwpt(); i < j; ++i) {
			int x;
			{
				double dist;
				if (!m_profile.size()) {
					continue;
				} else if (i >= m_profile.size()) {
					if (!m_profile.back().size())
						continue;
					dist = m_profile.back().back().get_dist();
				} else if (!m_profile[i].size()) {
					continue;
				} else {
					dist = m_profile[i].front().get_dist();
				}
				x = alt_width + Point::round<int,double>((dist - m_origindist) * m_scaledist);
			}
			if (x < alt_width || x > width)
				continue;
			Glib::ustring txt((*m_route)[i].get_icao_name());
			if (txt.empty())
				continue;
			Cairo::TextExtents ext;
			cr->get_text_extents(txt, ext);
			cr->move_to(x - ext.width / 2U - ext.x_bearing, height - 2U * dist_line_height + 2U - ext.y_bearing);
			cr->show_text(txt);
			cr->begin_new_path();
			cr->move_to(x - 2, height - 2 * dist_line_height);
			cr->rel_line_to(4, 0);
			cr->rel_line_to(-2, -4);
			cr->close_path();
			cr->fill();
		}
	}
	// normalize and draw terrain
	cr->save();
	Cairo::Matrix mnorm;
	cr->get_matrix(mnorm);
	cr->rectangle(alt_width, 0, width - alt_width, height - dist_height);
	cr->clip();
	set_color_sky(cr);
	cr->paint();
	cr->set_line_width(2);
	cr->translate(alt_width, height - dist_height);
	cr->scale(m_scaledist, -m_scaleelev);
	cr->translate(-m_origindist, -m_originelev);
	Cairo::Matrix mscaled;
	cr->get_matrix(mscaled);
	Cairo::Matrix mscaledmeter;
	cr->scale(1., Point::m_to_ft_dbl);
	cr->get_matrix(mscaledmeter);
	// draw max elev
	{
		cr->begin_new_path();
		cr->move_to(0, 0);
		double dist(0);
		for (profile_t::const_iterator pi(m_profile.begin()), pe(m_profile.end()); pi != pe; ++pi)
			for (Engine::ElevationProfileResult::Profile::const_iterator ppi(pi->begin()), ppe(pi->end()); ppi != ppe; ++ppi)
				cr->line_to(dist = ppi->get_dist(), ppi->get_maxelev());
		cr->line_to(dist, 0);
		cr->close_path();
	}
	cr->set_matrix(mnorm);
	cr->set_source_rgb(1.0, 1.0, 1.0);
	cr->fill_preserve();
	set_color_terrain_fill(cr);
	cr->fill_preserve();
	set_color_terrain(cr);
	cr->stroke();
	cr->set_matrix(mscaledmeter);
	// draw elev
	{
		cr->begin_new_path();
		cr->move_to(0, 0);
		double dist(0);
		for (profile_t::const_iterator pi(m_profile.begin()), pe(m_profile.end()); pi != pe; ++pi)
			for (Engine::ElevationProfileResult::Profile::const_iterator ppi(pi->begin()), ppe(pi->end()); ppi != ppe; ++ppi)
				cr->line_to(dist = ppi->get_dist(), ppi->get_elev());
		cr->line_to(dist, 0);
		cr->close_path();
	}
	cr->set_matrix(mnorm);
	set_color_terrain_fill(cr);
	cr->fill_preserve();
	set_color_terrain(cr);
	cr->stroke();
	cr->restore();
	// draw scales
	set_color_marking(cr);
	cr->set_line_width(1);
	for (;; elevstart += elevinc) {
		int y(height - dist_height - Point::round<int,double>((elevstart - m_originelev) * m_scaleelev));
		if (y < 0)
			break;
		cr->move_to(alt_width, y);
		cr->line_to(width, y);
		cr->stroke();
		std::ostringstream oss;
		oss << std::fixed << std::setprecision(0) << elevstart;
		Cairo::TextExtents ext;
		cr->get_text_extents(oss.str(), ext);
		cr->move_to(alt_width - 2 - ext.width - ext.x_bearing, y - ext.height / 2 - ext.y_bearing);
		cr->show_text(oss.str());
	}
	for (;; diststart += distinc) {
		int x(alt_width + Point::round<int,double>((diststart - m_origindist) * m_scaledist));
		if (x >= width)
			break;
		cr->move_to(x, 0);
		cr->line_to(x, height - dist_height);
		cr->stroke();
		std::ostringstream oss;
		oss << std::fixed << std::setprecision(0) << diststart;
		Cairo::TextExtents ext;
		cr->get_text_extents(oss.str(), ext);
		cr->move_to(x - ext.width / 2U - ext.x_bearing, height - dist_line_height + 2U - ext.y_bearing);
		cr->show_text(oss.str());
	}
	cr->restore();
	// normalize and draw route
	if (m_route) {
		cr->save();
		cr->rectangle(alt_width, 0, width - alt_width, height - dist_height);
		cr->clip();
		set_color_fplan(cr);
		cr->set_line_width(2);
		cr->set_matrix(mscaled);
		bool first(true);
		cr->begin_new_path();
		for (unsigned int i = 0, j = m_route->get_nrwpt(); i < j; ++i) {
			double dist;
			{
				if (!m_profile.size()) {
					continue;
				} else if (i >= m_profile.size()) {
					if (!m_profile.back().size())
						continue;
					dist = m_profile.back().back().get_dist();
				} else if (!m_profile[i].size()) {
					continue;
				} else {
					dist = m_profile[i].front().get_dist();
				}
			}
			float alt((*m_route)[i].get_true_altitude());
			if (first)
				cr->move_to(dist, alt);
			else
				cr->line_to(dist, alt);
			first = false;
		}
		cr->set_matrix(mnorm);
		cr->stroke();
		cr->restore();
	}
	// end
	cr->restore();
	return true;
}

void FlightDeckWindow::RouteProfileDrawingArea::on_size_allocate(Gtk::Allocation& allocation)
{
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        Gtk::DrawingArea::on_size_allocate(allocation);
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
}

void FlightDeckWindow::RouteProfileDrawingArea::on_show(void)
{
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        Gtk::DrawingArea::on_show();
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
	if (m_profile.empty())
	        async_startquery();
}

void FlightDeckWindow::RouteProfileDrawingArea::on_hide(void)
{
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        Gtk::DrawingArea::on_hide();
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
	m_dragx = m_dragy = -1;
}

bool FlightDeckWindow::RouteProfileDrawingArea::on_button_press_event(GdkEventButton *event)
{
	if (event->button == 1) {
		m_dragx = event->x;
		m_dragy = event->y;
	}
        if (true)
                std::cerr << "routeprofile button_press " << event->x << ' ' << event->y << ' ' << event->button << std::endl;
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        if (Gtk::DrawingArea::on_button_press_event(event))
                return true;
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        return false;
}

bool FlightDeckWindow::RouteProfileDrawingArea::on_button_release_event(GdkEventButton *event)
{
	m_dragx = m_dragy = -1;
        if (true)
                std::cerr << "routeprofile button_release " << event->x << ' ' << event->y << ' ' << event->button << std::endl;
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        if (Gtk::DrawingArea::on_button_release_event(event))
                return true;
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        return false;
}

bool FlightDeckWindow::RouteProfileDrawingArea::on_motion_notify_event(GdkEventMotion *event)
{
	if (m_dragx >= 0 && m_dragy >= 0) {
		int dx(event->x - m_dragx), dy(event->y - m_dragy);
		m_dragx = event->x;
		m_dragy = event->y;
		m_origindist -= dx / m_scaledist;
		m_originelev += dy / m_scaleelev;
		if (m_profile.size() && m_profile.back().size()) {
			double dmax(m_profile.back().back().get_dist());
			Gtk::Allocation allocation = get_allocation();
			dmax -= (allocation.get_width() - 50) / m_scaledist;
			if (m_origindist > dmax)
				m_origindist = dmax;
		}
		if (m_origindist < 0)
			m_origindist = 0;
		if (m_originelev > 60000)
			m_originelev = 60000;
		if (m_originelev < 0)
			m_originelev = 0;
		queue_draw();
	}
        if (false)
                std::cerr << "routeprofile motion_notify " << event->x << ' ' << event->y << std::endl;
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        if (Gtk::DrawingArea::on_motion_notify_event(event))
                return true;
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        return false;
}

void FlightDeckWindow::RouteProfileDrawingArea::async_cancel(void)
{
        Engine::ElevationProfileResult::cancel(m_profilequery);
}

void FlightDeckWindow::RouteProfileDrawingArea::async_clear(void)
{
	m_profilequery = Glib::RefPtr<Engine::ElevationProfileResult>();
}

void FlightDeckWindow::RouteProfileDrawingArea::async_connectdone(void)
{
        if (m_profilequery)
                m_profilequery->connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::async_done));
}

void FlightDeckWindow::RouteProfileDrawingArea::async_done(void)
{
	m_dispatch();
}

void FlightDeckWindow::RouteProfileDrawingArea::async_result(void)
{
	if (!m_profilequery || !m_profilequery->is_done())
		return;
	if (m_profilequery->is_error()) {
		m_profilequery = Glib::RefPtr<Engine::ElevationProfileResult>();
		return;
	}
	m_profile.push_back(m_profilequery->get_result());
	m_profilequery = Glib::RefPtr<Engine::ElevationProfileResult>();
	if (!m_route || !m_engine || m_profile.size() >= m_route->get_nrwpt() || !m_profile.back().size()) {
		m_profile.clear();
		return;
	}
	if (m_profile.size() > 1) {
		unsigned int idx(m_profile.size() - 2U);
		m_profile.back().adddist(m_profile[idx].front().get_dist() + 
					 (*m_route)[idx].get_coord().spheric_distance_nmi_dbl((*m_route)[idx + 1U].get_coord()));
	}
	if (get_is_drawable())
		queue_draw();
	{
		unsigned int idx(m_profile.size());
		if (idx + 1U < m_route->get_nrwpt()) {
			async_cancel();
			async_clear();
			m_profilequery = m_engine->async_elevation_profile((*m_route)[idx].get_coord(), (*m_route)[idx + 1U].get_coord(), 5.);
			async_connectdone();
		}
	}
}
